function DoAUVLocalization

% This code simulates an autonomous underwater vehicle (AUV) as it moves
% through the water. On the seabed are a number of sonar beacons at known
% locations. The AUV can send a sonar pulse to these beacons. Upon
% receiving this message, the beacons will immediately respond. The AUV
% measures the time-till-response for each of the beacons. Using these time
% measurements and the known locations for the beacons, the 3d location of
% the AUV can be determined.


nSteps = 100; %number of timesteps in the simulation

% position of beacons X Y Z (Z=0 is seabed)
Beacons  = [-40 -40 0; 
    40 -40 0;
    40 40 0;
    -40  40 0
    0  0  0]';

nBeacons = size(Beacons,2);

XvTrue = [-30,30,100]'; %true initial position of the AUV

Xv = XvTrue+10*randn(3,1); % initial position guess has some noise on it

c = 1500; %m/s speed of sound in water

RTrue  = [0.0008].^2; %actual error in sensor (for simulation)
R = [0.0008].^2; % assumed time error on measurements


% set up some graphics
figure(1);
clf; %clear the figure
plot3(Beacons(1,:),Beacons(2,:),Beacons(3,:),'kh');
hold on

for i = 1:nSteps
    
    % Simulate the ground truth motion of the vehicle
    % Vehicle descends through the water in a spiral
    dZ = -0.5;
    Vdt = 3;
    XvTrue = XvTrue + [ Vdt*cos(i/nSteps*2*pi);
                        Vdt*sin(i/nSteps*2*pi);
                        dZ];
    
    %%%% Your code here
    
    % First, get observations for each beacon using SimulateObservation.
    % You will have to implement ObsModelH first.
    
    % Then, using these observations and the known beacon positions
    % estimate the 3d position of the AUV using a non-linear least squares
    % method. The the first timestep the initial guess for the vehicle's
    % position is Xv. For later timesteps just use the solution from the
    % last timestep.
                    
                    
                    
    
    % Drawing
    plot3(XvTrue(1),XvTrue(2),XvTrue(3),'g.');
    plot3(Xv(1),Xv(2),Xv(3),'r+');
    axis tight;
    axis equal
    grid on
    drawnow
end;



%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
function z = ObsModelH(Xv,BeaconPosition,c)
% Implement the model for the measurement.
% The AUV measures the time beween sending a sonar message to the beacon
% and hearing the response.
% Given the vehicle position and the beacon position
% return time of flight between vehicle to beacon and back


%%%% Your code here


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
function  z = SimulateObservation(XvTrue,BeaconPosition,RTrue,c)
 % This function simulates the actual measurement made which is corrupted
 % by noise.

ground_truth_measurement = ObsModelH(XvTrue,BeaconPosition,c);
measurement_noise=sqrt(RTrue)*randn(1);

z = ground_truth_measurement + measurement_noise;




%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
function PlotEllipse(x,P,nSigma,z)
% used to plot 2d covariance ellipse of uncertainty in the xy plane at height z
% x is the mean (2x1 vector)
% P is the covariance matrix (2x2 Matrix)
% nSigma is the number of standard deviantions (3 is typical)

% This function starts with a unit circle and then rotates and scales it to
% give the desired ellipsoid.
eH = [];
P = P(1:2,1:2);
x = x(1:2);
if(~any(diag(P)==0))
    [V,D] = eig(P);
    y = nSigma*[cos(0:0.1:2*pi);sin(0:0.1:2*pi)];
    el = V*sqrtm(D)*y;
    el = [el el(:,1)]+repmat(x,1,size(el,2)+1);
    line(el(1,:),el(2,:),z*ones(length(el(1,:)),1 ) );
end;



%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
function DrawCovarianceEllipsoid(x,P,nSigma)
% used to plot 3d covariance ellipsoids of uncertainty
% x is the mean (3x1 vector)
% P is the covariance matrix (3x3 Matrix)
% nSigma is the number of standard deviantions (3 is typical)

% This function starts with a unit sphere and then rotates and scales it to
% give the desired ellipsoid.


[X,Y,Z] = sphere(20);

M = [reshape(X,1,prod(size(X)));
    reshape(Y,1,prod(size(Y)));
    reshape(Z,1,prod(size(Z)))];

[V,D] = eig(P(1:3,1:3));

%rotate by V to align graphics,
%multiply by root(D) to scale.
M = nSigma*V*sqrtm(abs(D))*M;

X = reshape(M(1,:),size(X))+x(1);
Y = reshape(M(2,:),size(Y))+x(2);
Z = reshape(M(3,:),size(Z))+x(3);

surf(X,Y,Z,'facealpha',0.1,'edgealpha', 0.1,'facecolor','r','EdgeColor','none');
